from __future__ import annotations

import numpy as np


def equidistant_angles(n_points: int) -> np.ndarray:
    """
    An array of `n_points` equidistantly spaced angles in radians
    """
    max_angle = 2.0 * np.pi * (n_points - 1.0) / n_points
    return np.linspace(0, max_angle, n_points)


def points_on_circle(
    n_points: int, radius: float, include_centre: bool = False
) -> np.ndarray:
    points = [
        [radius * np.sin(angle), radius * np.cos(angle)]
        for angle in equidistant_angles(n_points)
    ]
    if include_centre:
        points.append([0.0, 0.0])
    return np.array(points)


def rotate_point(
    point: tuple[float, float],
    angle_radians: float,
    pivot_point: tuple[float, float] = (0, 0),
) -> tuple[float, float]:
    """
    Rotate `point` by an angle `angle_radians` around the point `pivot_point`
    """
    s = np.sin(angle_radians)
    c = np.cos(angle_radians)
    x = point[0] - pivot_point[0]
    y = point[1] - pivot_point[1]
    return pivot_point[0] + c * x - s * y, pivot_point[1] + s * x + c * y


class PointRotator:
    """
    Rotate a point (x,y) by a fixed angle in degrees around the origin
    """

    def __init__(self, angle_degrees: float):
        self.angle_degrees = angle_degrees
        angle_radians = angle_degrees * np.pi / 180.0
        self._s = np.sin(angle_radians)
        self._c = np.cos(angle_radians)

    def __call__(self, point: tuple[float, float]) -> np.ndarray:
        return np.array(
            [
                self._c * point[0] - self._s * point[1],
                self._s * point[0] + self._c * point[1],
            ]
        )


class JoystickPointUpdater:
    """
    Update a point (x,y) according to a velocity vector (vx, vy), with optional rotation of the vector

    vx, vy should lie in the range [-1, 1]

    The returned point is given by

    x -> x + `max_speed` * vx'
    y -> y + `max_speed` * vy'

    Where (vx',vy') is (vx, vy) rotated by `angle_degrees`

    The returned point is clipped to lie within the screen,
    where the screen has height 1 and (0,0) lies in the centre of the screen
    So the y-range is always [-0.5, 0.5], and the x-range is this multiplied by (width/height)
    """

    def __init__(
        self,
        angle_degrees: float,
        max_speed: float,
        window_size: np.ndarray | None = None,
    ):
        self.angle_degrees = angle_degrees
        self.max_speed = max_speed
        if window_size is not None and window_size.shape[-1] >= 2:
            self.clip_window = 0.5 * np.array([window_size[0] / window_size[1], 1.0])
        else:
            self.clip_window = np.array([0.5, 0.5])
        angle_radians = angle_degrees * np.pi / 180.0
        self._s = self.max_speed * np.sin(angle_radians)
        self._c = self.max_speed * np.cos(angle_radians)

    def __call__(self, point: np.ndarray, velocity: tuple[float, float]) -> np.ndarray:
        return np.clip(
            np.array(
                [
                    point[0] + self._c * velocity[0] + self._s * velocity[1],
                    point[1] + self._s * velocity[0] - self._c * velocity[1],
                ]
            ),
            -self.clip_window,
            self.clip_window,
        )


def to_target_dists(
    pos: np.ndarray, target_xys: np.ndarray, target_index: int, has_central_target: bool
) -> tuple[float, float]:
    rms_dists = np.linalg.norm(target_xys - np.array(pos), axis=1)
    n_targets_excluding_central = target_xys.shape[0]
    if has_central_target:
        # exclude central target from "distance to any target" measure
        n_targets_excluding_central = n_targets_excluding_central - 1
    # dist to correct target, min distance to any target (excluding center target)
    return rms_dists[target_index], np.min(rms_dists[:n_targets_excluding_central])



from __future__ import annotations

from psychopy.hardware import joystick
from psychopy.visual.window import Window

from vstt import joystick_wrapper


def test_joystick_wrapper(window: Window) -> None:
    # check that calling repeatedly doesn't raise an exception
    if joystick.getNumJoysticks() > 0:
        assert joystick_wrapper.get_joystick() is not None
        assert joystick_wrapper.get_joystick() is not None
    else:
        assert joystick_wrapper.get_joystick() is None
        assert joystick_wrapper.get_joystick() is None



from __future__ import annotations

import copy
from typing import Any
from typing import Mapping

import numpy as np
from psychopy.gui.qtgui import DlgFromDict

from vstt.common import import_typed_dict
from vstt.vtypes import Trial


def describe_trial(trial: Trial) -> str:
    repeats = f"{trial['weight']} repeat{'s' if trial['weight'] > 1 else ''}"
    targets = f"target{'s' if trial['num_targets'] > 1 else ''}"
    if trial["condition_timeout"] > 0.0:
        repeats = f"up to {repeats}"
        targets = f"{targets} in {trial['condition_timeout']}s"
    return f"{repeats} of {trial['num_targets']} {trial['target_order']} {targets}"


def describe_trials(trials: list[Trial]) -> str:
    return "\n".join(["  - " + describe_trial(trial) for trial in trials])


def default_trial() -> Trial:
    return {
        "weight": 1,
        "condition_timeout": 0.0,
        "num_targets": 8,
        "target_order": "clockwise",
        "target_indices": "0 1 2 3 4 5 6 7",
        "add_central_target": True,
        "hide_target_when_reached": True,
        "turn_target_to_green_when_reached": False,
        "show_target_labels": False,
        "target_labels": "0 1 2 3 4 5 6 7",
        "fixed_target_intervals": False,
        "target_duration": 5.0,
        "central_target_duration": 5.0,
        "pre_target_delay": 0.0,
        "pre_central_target_delay": 0.0,
        "pre_first_target_extra_delay": 0.0,
        "target_distance": 0.4,
        "target_size": 0.04,
        "central_target_size": 0.02,
        "show_inactive_targets": True,
        "ignore_incorrect_targets": True,
        "play_sound": True,
        "use_joystick": False,
        "joystick_max_speed": 0.02,
        "show_cursor": True,
        "cursor_size": 0.02,
        "show_cursor_path": True,
        "automove_cursor_to_center": False,
        "freeze_cursor_between_targets": False,
        "cursor_rotation_degrees": 0.0,
        "post_trial_delay": 0.0,
        "post_trial_display_results": False,
        "post_block_delay": 10.0,
        "post_block_display_results": True,
        "show_delay_countdown": True,
        "enter_to_skip_delay": True,
    }


def trial_labels() -> dict:
    return {
        "weight": "Repetitions",
        "condition_timeout": "Maximum time, 0=unlimited (secs)",
        "num_targets": "Number of targets",
        "target_order": "Target order",
        "target_indices": "Target indices",
        "add_central_target": "Add a central target",
        "hide_target_when_reached": "Hide target when reached",
        "turn_target_to_green_when_reached": "Turn target to green when reached",
        "show_target_labels": "Display target labels",
        "target_labels": "Target labels",
        "fixed_target_intervals": "Fixed target display intervals",
        "target_duration": "Target display duration (secs)",
        "central_target_duration": "Central target display duration (secs)",
        "pre_target_delay": "Delay before outer target display (secs)",
        "pre_central_target_delay": "Delay before central target display (secs)",
        "pre_first_target_extra_delay": "Extra delay before first outer target of condition (secs)",
        "target_distance": "Distance to targets (screen height fraction)",
        "target_size": "Target size (screen height fraction)",
        "central_target_size": "Central target size (screen height fraction)",
        "show_inactive_targets": "Show inactive targets",
        "ignore_incorrect_targets": "Ignore cursor hitting incorrect target",
        "play_sound": "Play a sound on target display",
        "use_joystick": "Control the cursor using a joystick",
        "joystick_max_speed": "Maximum joystick speed (screen height fraction per frame)",
        "show_cursor": "Show cursor",
        "cursor_size": "Cursor size (screen height fraction)",
        "show_cursor_path": "Show cursor path",
        "automove_cursor_to_center": "Automatically move cursor to center",
        "freeze_cursor_between_targets": "Freeze cursor until target is displayed",
        "cursor_rotation_degrees": "Cursor rotation (degrees)",
        "post_trial_delay": "Delay between trials (secs)",
        "post_trial_display_results": "Display results after each trial",
        "post_block_delay": "Delay after last trial (secs)",
        "post_block_display_results": "Display combined results after last trial",
        "show_delay_countdown": "Display a countdown during delays",
        "enter_to_skip_delay": "Skip delay by pressing enter key",
    }


def get_trial_from_user(
    initial_trial: Trial | None = None,
) -> Trial | None:
    trial = copy.deepcopy(initial_trial) if initial_trial else default_trial()
    order_of_targets = [trial["target_order"]]
    for target_order in ["clockwise", "anti-clockwise", "random", "fixed"]:
        if target_order != order_of_targets[0]:
            order_of_targets.append(target_order)
    trial["target_order"] = order_of_targets
    dialog = DlgFromDict(
        trial, title="Trial conditions", labels=trial_labels(), sortKeys=False
    )
    if not dialog.OK:
        return None
    return import_and_validate_trial(trial)


def import_and_validate_trial(trial_or_dict: Mapping[str, Any]) -> Trial:
    trial = import_typed_dict(trial_or_dict, default_trial())
    # make any negative time durations zero
    for duration in [
        "condition_timeout",
        "target_duration",
        "central_target_duration",
        "pre_target_delay",
        "pre_central_target_delay",
        "pre_first_target_extra_delay",
        "post_trial_delay",
        "post_block_delay",
    ]:
        if trial[duration] < 0.0:  # type: ignore
            trial[duration] = 0.0  # type: ignore
    # convert string of target indices to a numpy array of ints
    target_indices = np.fromstring(trial["target_indices"], dtype="int", sep=" ")
    if trial["target_order"] == "fixed":
        # clip indices to valid range
        target_indices = np.clip(target_indices, 0, trial["num_targets"] - 1)
    else:
        # construct clockwise sequence
        target_indices = np.array(range(trial["num_targets"]))
        if trial["target_order"] == "anti-clockwise":
            target_indices = np.flip(target_indices)
        elif trial["target_order"] == "random":
            rng = np.random.default_rng()
            rng.shuffle(target_indices)
    # convert array of indices back to a string
    trial["target_indices"] = " ".join(f"{int(i)}" for i in target_indices)
    return trial



from __future__ import annotations

import pathlib
from typing import Any

import qt_test_utils as qtu
from pytest import MonkeyPatch
from qtpy.QtWidgets import QFileDialog
from qtpy.QtWidgets import QInputDialog

import vstt
from vstt.experiment import Experiment
from vstt.gui import Gui


def test_gui_run_no_trials() -> None:
    gui = Gui(filename=None)
    # remove trial conditions
    qtu.press_up_key(gui.trials_widget._widget_list_trials)
    qtu.click(gui.trials_widget._btn_remove)
    # mwt to press escape ok when dialog appears
    mwt = qtu.ModalWidgetTimer(["Enter"])
    mwt.start()
    # try to run experiment
    open_action = gui.toolbar.actions()[4]
    assert open_action.text() == "&Run"
    open_action.trigger()
    assert mwt.widget_type == "QMessageBox"
    assert (
        mwt.widget_text == "Please add a trial condition before running the experiment."
    )


def test_gui_new_file() -> None:
    gui = Gui(filename=None)
    # initially uses default metadata, display options and trials - no results, no unsaved changes
    assert gui.experiment.has_unsaved_changes is False
    assert "*" not in gui.windowTitle()
    # make a change so the experiment has unsaved changes
    qtu.click(gui.display_options_widget._widgets["to_target_paths"])
    assert gui.experiment.has_unsaved_changes is True
    assert "*" in gui.windowTitle()
    # new file
    # mwt to press no when save changes dialog appears
    mwt = qtu.ModalWidgetTimer(["N"])
    mwt.start()
    new_action = gui.toolbar.actions()[0]
    assert new_action.text() == "&New"
    new_action.trigger()
    assert mwt.widget_type == "QMessageBox"
    assert mwt.widget_text == "Save your changes before continuing?"
    assert gui.experiment.has_unsaved_changes is False
    assert "*" not in gui.windowTitle()


def test_gui_no_file() -> None:
    gui = Gui(filename=None)
    # initially uses default metadata, display options and trials - no results, no unsaved changes
    assert gui.experiment.metadata == vstt.meta.default_metadata()
    assert gui.experiment.display_options == vstt.display.default_display_options()
    assert gui.experiment.trial_list == [vstt.trial.default_trial()]
    assert gui.experiment.trial_handler_with_results is None
    assert gui.experiment.has_unsaved_changes is False
    assert "*" not in gui.windowTitle()
    # file -> open -> cancel: does nothing
    # mwt to press escape when file open dialog appears
    mwt = qtu.ModalWidgetTimer(["Escape"])
    mwt.start()
    open_action = gui.toolbar.actions()[1]
    assert open_action.text() == "&Open"
    open_action.trigger()
    assert mwt.widget_type == "QFileDialog.Open"
    assert mwt.widget_text == "Open an experiment"
    # make a change so the experiment has unsaved changes
    qtu.click(gui.display_options_widget._widgets["to_target_paths"])
    assert gui.experiment.has_unsaved_changes is True
    assert "*" in gui.windowTitle()
    # mwt to click yes when asked if we want to save our changes
    mwt1 = qtu.ModalWidgetTimer(["Y"])
    # mwt to press escape at subsequent file save dialog
    mwt2 = qtu.ModalWidgetTimer(["Escape"], start_after=mwt1)
    mwt1.start()
    # try to close the gui: yes to save changes but then escape at file save dialog: gui not closed
    gui.close()
    assert mwt1.widget_type == "QMessageBox"
    assert mwt1.widget_text == "Save your changes before continuing?"
    assert mwt2.widget_type == "QFileDialog.Save"
    assert mwt2.widget_text == "Save experiment"
    # mwt to click no when asked if we want to save our changes
    mwt = qtu.ModalWidgetTimer(["N"])
    mwt.start()
    # try close the gui: no to save changes: gui closes
    gui.close()
    assert mwt.widget_type == "QMessageBox"
    assert mwt.widget_text == "Save your changes before continuing?"


def test_gui_file_with_results(
    tmp_path: pathlib.Path, experiment_with_results: Experiment
) -> None:
    filename = str(tmp_path / "experiment.psydat")
    experiment_with_results.save_psydat(filename)
    gui = Gui(filename=filename)
    assert gui.experiment.metadata == experiment_with_results.metadata
    assert gui.experiment.display_options == experiment_with_results.display_options
    assert len(gui.experiment.trial_list) == len(experiment_with_results.trial_list)
    assert gui.experiment.has_unsaved_changes is False
    assert gui.experiment.trial_handler_with_results is not None
    assert (
        gui.results_widget._list_trials.count()
        == gui.experiment.trial_handler_with_results.nTotal
    )
    # close the gui: no modal prompt as there are no unsaved changes
    gui.close()


def test_gui_import_excel(
    tmp_path: pathlib.Path, experiment_with_results: Experiment
) -> None:
    filepath = tmp_path / "experiment.xlsx"
    experiment_with_results.save_excel(str(filepath), "trial")
    gui = Gui(filename=str(filepath))
    assert gui.experiment.metadata == experiment_with_results.metadata
    assert gui.experiment.display_options == experiment_with_results.display_options
    assert len(gui.experiment.trial_list) == len(experiment_with_results.trial_list)
    assert gui.experiment.has_unsaved_changes is True
    assert gui.experiment.trial_handler_with_results is None
    assert "*" in gui.windowTitle()
    assert gui.experiment.filename == str(filepath.with_suffix(".psydat"))


def test_gui_import_json(
    tmp_path: pathlib.Path, experiment_with_results: Experiment
) -> None:
    filepath = tmp_path / "experiment.json"
    experiment_with_results.save_json(str(filepath))
    gui = Gui(filename=str(filepath))
    assert gui.experiment.metadata == experiment_with_results.metadata
    assert gui.experiment.display_options == experiment_with_results.display_options
    assert len(gui.experiment.trial_list) == len(experiment_with_results.trial_list)
    assert gui.experiment.has_unsaved_changes is True
    assert gui.experiment.trial_handler_with_results is None
    assert "*" in gui.windowTitle()
    assert gui.experiment.filename == str(filepath.with_suffix(".psydat"))


def test_gui_import_export(
    tmp_path: pathlib.Path,
    experiment_with_results: Experiment,
    monkeypatch: MonkeyPatch,
) -> None:
    filepath = tmp_path / "experiment.psydat"
    experiment_with_results.save_psydat(str(filepath))
    for output_file, filter, data_option in [
        ("t1.xlsx", "Excel file (*.xlsx)", "One page for each trial (default)"),
        ("t2.xlsx", "Excel file (*.xlsx)", "One page for each target"),
        ("t3.json", "JSON file (*.json)", ""),
    ]:
        export_filepath = filepath.parent / output_file
        assert not export_filepath.is_file()

        # monkey patch QFileDialog.getSaveFileName to just return desired filename
        def mock_get_save_file_name(
            *args: Any,
            mock_return_value: tuple[str, str] = (str(export_filepath), filter),
            **kwargs: Any,
        ) -> tuple[str, str]:
            return mock_return_value

        monkeypatch.setattr(QFileDialog, "getSaveFileName", mock_get_save_file_name)

        # monkey patch QInputDialog.getItem to just return desired option
        def mock_get_item(
            *args: Any,
            mock_return_value: tuple[str, bool] = (data_option, True),
            **kwargs: Any,
        ) -> tuple[str, bool]:
            return mock_return_value

        monkeypatch.setattr(QInputDialog, "getItem", mock_get_item)

        gui = Gui(filename=str(filepath))
        assert gui.experiment.metadata == experiment_with_results.metadata
        assert gui.experiment.display_options == experiment_with_results.display_options
        assert len(gui.experiment.trial_list) == len(experiment_with_results.trial_list)
        assert gui.experiment.has_unsaved_changes is False
        export_action = gui.toolbar.actions()[3]
        assert export_action.text() == "&Export"
        export_action.trigger()
        assert export_filepath.is_file()
        # re-import exported file
        exp = Experiment(str(export_filepath))
        assert exp.metadata == experiment_with_results.metadata
        assert exp.display_options == experiment_with_results.display_options
        assert exp.trial_list == experiment_with_results.trial_list
        assert exp.trial_handler_with_results is None
        assert exp.has_unsaved_changes is True
        assert exp.filename == str(export_filepath.with_suffix(".psydat"))


def test_gui_invalid_file(tmp_path: pathlib.Path) -> None:
    for extension in ["psydat", "xlsx", "json", "zzz"]:
        file = tmp_path / f"invalid.{extension}"
        file.write_text("beep boop")
        # mwt to click ok when failed to open file message box is shown
        mwt = qtu.ModalWidgetTimer(["Enter"])
        mwt.start()
        gui = Gui(filename=str(file))
        assert mwt.widget_type == "QMessageBox"
        assert "Could not read file" in mwt.widget_text
        assert extension in mwt.widget_text
        gui.close()


def test_gui_nonexistent_file() -> None:
    # mwt to click ok when failed to open file message box is shown
    mwt = qtu.ModalWidgetTimer(["Enter"])
    mwt.start()
    gui = Gui(filename="I dont exist")
    assert mwt.widget_type == "QMessageBox"
    assert "Could not read file" in mwt.widget_text
    gui.close()



from __future__ import annotations

from psychopy.visual.window import Window
from qtpy import QtCore
from qtpy import QtWidgets

from vstt.experiment import Experiment
from vstt.trial import Trial
from vstt.trial import describe_trial
from vstt.trial import get_trial_from_user


class TrialsWidget(QtWidgets.QWidget):
    experiment_modified = QtCore.Signal()

    def __init__(
        self, parent: QtWidgets.QWidget | None = None, win: Window | None = None
    ):
        super().__init__(parent)
        self._win = win
        self._experiment = Experiment()

        outer_layout = QtWidgets.QVBoxLayout()
        group_box = QtWidgets.QGroupBox("Trial Conditions")
        outer_layout.addWidget(group_box)
        inner_layout = QtWidgets.QGridLayout()
        group_box.setLayout(inner_layout)
        self._widget_list_trials = QtWidgets.QListWidget()
        self._widget_list_trials.currentRowChanged.connect(self._row_changed)
        self._widget_list_trials.itemDoubleClicked.connect(self._edit_trial)
        inner_layout.addWidget(self._widget_list_trials, 0, 0, 1, 5)
        self._btn_add = QtWidgets.QPushButton("Add")
        self._btn_add.clicked.connect(self._btn_add_clicked)
        inner_layout.addWidget(self._btn_add, 1, 0)
        self._btn_edit = QtWidgets.QPushButton("Edit")
        self._btn_edit.clicked.connect(self._edit_trial)
        inner_layout.addWidget(self._btn_edit, 1, 1)
        self._btn_move_up = QtWidgets.QPushButton("Move Up")
        self._btn_move_up.clicked.connect(self._btn_move_up_clicked)
        inner_layout.addWidget(self._btn_move_up, 1, 2)
        self._btn_move_down = QtWidgets.QPushButton("Move Down")
        self._btn_move_down.clicked.connect(self._btn_move_down_clicked)
        inner_layout.addWidget(self._btn_move_down, 1, 3)
        self._btn_remove = QtWidgets.QPushButton("Remove")
        self._btn_remove.clicked.connect(self._btn_remove_clicked)
        inner_layout.addWidget(self._btn_remove, 1, 4)
        self.setLayout(outer_layout)
        self._row_changed()

    def _is_valid(self, row: int) -> bool:
        return 0 <= row < len(self._experiment.trial_list)

    def _can_move_up(self, row: int) -> bool:
        return 1 <= row < len(self._experiment.trial_list)

    def _can_move_down(self, row: int) -> bool:
        return 0 <= row < len(self._experiment.trial_list) - 1

    def _row_changed(self) -> None:
        row = self._widget_list_trials.currentRow()
        self._btn_edit.setEnabled(self._is_valid(row))
        self._btn_remove.setEnabled(self._is_valid(row))
        self._btn_move_up.setEnabled(self._can_move_up(row))
        self._btn_move_down.setEnabled(self._can_move_down(row))

    def _current_trial(self) -> Trial | None:
        row = self._widget_list_trials.currentRow()
        if not self._is_valid(row):
            return None
        return self._experiment.trial_list[row]

    def _confirm_clear_data(self) -> bool:
        if self._experiment.trial_handler_with_results is None:
            self._experiment.has_unsaved_changes = True
            self.experiment_modified.emit()
            return True
        yes_no = QtWidgets.QMessageBox.question(
            self,
            "Clear existing results?",
            "Modifying the trial conditions will clear the existing results. Continue?",
        )
        if yes_no == QtWidgets.QMessageBox.Yes:
            self._experiment.clear_results()
            self._experiment.has_unsaved_changes = True
            self.experiment_modified.emit()
            return True
        return False

    def _btn_add_clicked(self) -> None:
        if not self._confirm_clear_data():
            return
        trial = get_trial_from_user(self._current_trial())
        if trial is not None:
            self._experiment.trial_list.append(trial)
            self._widget_list_trials.addItem(describe_trial(trial))
            self._widget_list_trials.setCurrentRow(len(self._experiment.trial_list) - 1)

    def _swap_rows(self, row1: int, row2: int) -> None:
        if not self._is_valid(row1) or not self._is_valid(row2):
            return
        if not self._confirm_clear_data():
            return
        self._experiment.trial_list[row1], self._experiment.trial_list[row2] = (
            self._experiment.trial_list[row2],
            self._experiment.trial_list[row1],
        )
        self._widget_list_trials.insertItem(
            row2, self._widget_list_trials.takeItem(row1)
        )
        self._widget_list_trials.setCurrentRow(row2)

    def _btn_move_up_clicked(self) -> None:
        row = self._widget_list_trials.currentRow()
        self._swap_rows(row, row - 1)

    def _btn_move_down_clicked(self) -> None:
        row = self._widget_list_trials.currentRow()
        self._swap_rows(row, row + 1)

    def _btn_remove_clicked(self) -> None:
        row = self._widget_list_trials.currentRow()
        if not self._is_valid(row):
            return
        if not self._confirm_clear_data():
            return
        self._experiment.trial_list.pop(row)
        self._widget_list_trials.takeItem(row)
        self._row_changed()

    def _edit_trial(self) -> None:
        row = self._widget_list_trials.currentRow()
        if not self._is_valid(row):
            return
        if not self._confirm_clear_data():
            return
        edited_trial = get_trial_from_user(self._experiment.trial_list[row])
        if edited_trial is not None:
            self._experiment.trial_list[row] = edited_trial
            item = self._widget_list_trials.item(row)
            assert item is not None
            item.setText(describe_trial(edited_trial))

    @property
    def experiment(self) -> Experiment:
        return self._experiment

    @experiment.setter
    def experiment(self, experiment: Experiment) -> None:
        self._experiment = experiment
        self._widget_list_trials.clear()
        self._widget_list_trials.addItems(
            [describe_trial(trial) for trial in experiment.trial_list]
        )



from __future__ import annotations

from pytest import approx

import vstt


def test_describe_trial() -> None:
    trial = vstt.trial.default_trial()
    assert vstt.trial.describe_trial(trial) == "1 repeat of 8 clockwise targets"
    trial["weight"] = 3
    trial["num_targets"] = 1
    trial["target_order"] = "fixed"
    assert vstt.trial.describe_trial(trial) == "3 repeats of 1 fixed target"
    trial["weight"] = 2
    trial["num_targets"] = 5
    trial["target_order"] = "random"
    assert vstt.trial.describe_trial(trial) == "2 repeats of 5 random targets"


def test_describe_trials() -> None:
    trials = [vstt.trial.default_trial()]
    trial = vstt.trial.default_trial()
    trial["weight"] = 3
    trial["num_targets"] = 1
    trial["target_order"] = "fixed"
    trials.append(trial)
    assert (
        vstt.trial.describe_trials(trials)
        == "  - 1 repeat of 8 clockwise targets\n  - 3 repeats of 1 fixed target"
    )


def test_default_trial() -> None:
    trial = vstt.trial.default_trial()
    assert len(trial) == len(vstt.trial.trial_labels())
    assert isinstance(trial["target_indices"], str)
    assert len(trial["target_indices"].split(" ")) == trial["num_targets"]


def test_trial_labels() -> None:
    trial = vstt.trial.default_trial()
    labels = vstt.trial.trial_labels()
    assert len(trial) == len(labels)
    assert trial.keys() == labels.keys()


def test_import_trial() -> None:
    default_trial = vstt.trial.default_trial()
    trial_dict = {
        "weight": 2,
        "condition_timeout": 0,
        "num_targets": 6,
        "target_order": "clockwise",
        "target_indices": "0 1 2 3 4 5",
        "add_central_target": True,
        "show_target_labels": False,
        "hide_target_when_reached": True,
        "turn_target_to_green_when_reached": False,
        "target_labels": "0 1 2 3 4 5",
        "fixed_target_intervals": False,
        "target_duration": 3,
        "central_target_duration": 3,
        "pre_target_delay": 0,
        "pre_central_target_delay": 0,
        "pre_first_target_extra_delay": 0,
        "target_distance": 0.3,
        "target_size": 0.03,
        "central_target_size": 0.01,
        "show_inactive_targets": False,
        "ignore_incorrect_targets": True,
        "play_sound": True,
        "use_joystick": True,
        "joystick_max_speed": 0.001,
        "show_cursor": False,
        "cursor_size": 0.0123,
        "show_cursor_path": True,
        "automove_cursor_to_center": True,
        "freeze_cursor_between_targets": True,
        "cursor_rotation_degrees": 45,
        "post_trial_delay": 0.2,
        "post_trial_display_results": True,
        "post_block_delay": 2.0,
        "post_block_display_results": False,
        "show_delay_countdown": False,
        "enter_to_skip_delay": True,
    }
    # start with a dict containing valid values for all keys
    for key in default_trial:
        assert key in trial_dict
    # all valid keys are imported
    trial = vstt.trial.import_and_validate_trial(trial_dict)
    for key in trial:
        assert trial[key] == trial_dict[key]  # type: ignore
    # if any keys are missing, default values are used instead
    missing_keys = [
        "weight",
        "cursor_rotation_degrees",
        "post_trial_delay",
        "post_block_display_results",
    ]
    for key in missing_keys:
        trial_dict.pop(key)
    # unknown keys are ignored
    trial_dict["unknown_key1"] = "ignore me"
    trial_dict["unknown_key2"] = False
    trial = vstt.trial.import_and_validate_trial(trial_dict)
    for key in trial:
        if key in missing_keys:
            assert trial[key] == default_trial[key]  # type: ignore
        else:
            assert trial[key] == trial_dict[key]  # type: ignore


def test_validate_trial_durations() -> None:
    trial = vstt.trial.default_trial()
    # positive durations are not modified
    trial["target_duration"] = 1
    trial["central_target_duration"] = 1
    trial["pre_target_delay"] = 0.1
    trial["pre_central_target_delay"] = 0.087
    trial["pre_first_target_extra_delay"] = 0.08123
    trial["post_trial_delay"] = 0.2
    trial["post_block_delay"] = 0.7
    vtrial = vstt.trial.import_and_validate_trial(trial)
    assert vtrial["target_duration"] == 1
    assert vtrial["central_target_duration"] == 1
    assert vtrial["pre_target_delay"] == approx(0.1)
    assert vtrial["pre_central_target_delay"] == approx(0.087)
    assert vtrial["pre_first_target_extra_delay"] == approx(0.08123)
    assert vtrial["post_trial_delay"] == approx(0.2)
    assert vtrial["post_block_delay"] == approx(0.7)
    # negative durations are cast to zero
    trial["target_duration"] = -1
    trial["central_target_duration"] = -0.8
    trial["pre_target_delay"] = -0.1
    trial["pre_central_target_delay"] = -0.087
    trial["pre_first_target_extra_delay"] = -0.08123
    trial["post_trial_delay"] = -0.2
    trial["post_block_delay"] = -0.7
    vtrial = vstt.trial.import_and_validate_trial(trial)
    assert vtrial["target_duration"] == 0
    assert vtrial["central_target_duration"] == 0
    assert vtrial["pre_target_delay"] == 0
    assert vtrial["pre_central_target_delay"] == 0
    assert vtrial["pre_first_target_extra_delay"] == 0
    assert vtrial["post_trial_delay"] == 0
    assert vtrial["post_block_delay"] == 0


def test_validate_trial_target_order() -> None:
    trial = vstt.trial.default_trial()
    assert isinstance(trial["target_indices"], str)
    # clockwise
    trial["target_order"] = "clockwise"
    vtrial = vstt.trial.import_and_validate_trial(trial)
    assert isinstance(vtrial["target_indices"], str)
    assert vtrial["target_indices"] == "0 1 2 3 4 5 6 7"
    # anti-clockwise
    trial["target_order"] = "anti-clockwise"
    vtrial = vstt.trial.import_and_validate_trial(trial)
    assert isinstance(vtrial["target_indices"], str)
    assert vtrial["target_indices"] == "7 6 5 4 3 2 1 0"
    # random
    trial["target_order"] = "random"
    vtrial = vstt.trial.import_and_validate_trial(trial)
    assert isinstance(vtrial["target_indices"], str)
    assert len(set(vtrial["target_indices"].split(" "))) == 8
    # fixed & valid
    trial["target_order"] = "fixed"
    trial["target_indices"] = "0 1 2 3 4 5 6 7"
    vtrial = vstt.trial.import_and_validate_trial(trial)
    assert isinstance(vtrial["target_indices"], str)
    assert vtrial["target_indices"] == "0 1 2 3 4 5 6 7"
    # fixed & invalid - clipped to nearest valid indices
    trial["target_order"] = "fixed"
    trial["target_indices"] = "-2 8 1 5 12 -5"
    vtrial = vstt.trial.import_and_validate_trial(trial)
    assert isinstance(vtrial["target_indices"], str)
    assert vtrial["target_indices"] == "0 7 1 5 7 0"



from __future__ import annotations

import logging
from typing import Any

import numpy as np
import pandas as pd
from numpy import linalg as LA
from psychopy.data import TrialHandlerExt
from psychopy.event import xydist
from shapely.geometry import LineString
from shapely.ops import polygonize
from shapely.ops import unary_union

min_distance: float = 1e-12


def list_dest_stat_label_units() -> list[tuple[str, list[tuple[str, str, str]]]]:
    list_dest_stats = []
    for destination in ["target", "center"]:
        stats = []
        for base_stat, label, unit in [
            ("reaction_time", "Reaction", "s"),
            ("movement_time", "Movement", "s"),
            ("time", "Time", "s"),
            ("distance", "Distance", ""),
            ("rmse", "RMSE", ""),
            ("success", "Success", ""),
            ("spatial_error", "Spatial", ""),
        ]:
            stats.append((f"to_{destination}_{base_stat}", label, unit))
        list_dest_stats.append((destination, stats))
    list_dest_stats.append(("", [("area", "Area", "")]))
    list_dest_stats.append(("", [("normalized_area", "Normalized Area", "")]))
    list_dest_stats.append(("", [("peak_velocity", "Peak Velocity", "")]))
    list_dest_stats.append(("", [("peak_acceleration", "Peak Acceleration", "")]))
    list_dest_stats.append(
        (
            "",
            [("movement_time_at_peak_velocity", "Movement Time at Peak Velocity", "s")],
        )
    )
    list_dest_stats.append(
        ("", [("total_time_at_peak_velocity", "Total Time at Peak Velocity", "s")])
    )
    list_dest_stats.append(
        (
            "",
            [
                (
                    "movement_distance_at_peak_velocity",
                    "Movement Distance at Peak Velocity",
                    "",
                )
            ],
        )
    )
    list_dest_stats.append(
        ("", [("rmse_movement_at_peak_velocity", "RMSE Movement at Peak Velocity", "")])
    )
    return list_dest_stats


def _get_trial_data_columns() -> list[str]:
    return [
        "i_trial",
        "i_rep",
        "i_target",
        "condition_index",
        "target_index",
        "target_pos",
        "target_radius",
        "to_target_timestamps",
        "to_target_mouse_positions",
        "to_target_success",
        "to_target_num_timestamps_before_visible",
        "center_pos",
        "center_radius",
        "to_center_timestamps",
        "to_center_mouse_positions",
        "to_center_success",
        "to_center_num_timestamps_before_visible",
    ]


def _get_dat(
    data: dict, key: str, index: tuple, i_target: int, default_value: Any
) -> Any:
    ar = data.get(key)
    if ar is None:
        logging.debug(
            f"Key '{key}' not found in data, using default value {default_value}"
        )
        return default_value
    try:
        return ar[index][i_target]
    except IndexError:
        logging.debug(
            f"Index error for key '{key}', index '{index}', i_target '{i_target}', using default value {default_value}"
        )
    return default_value


def _get_target_data(
    trial_handler: TrialHandlerExt, index: tuple, i_target: int
) -> list:
    data = trial_handler.data
    condition_index = trial_handler.sequenceIndices[index]
    conditions = trial_handler.trialList[condition_index]
    target_radius = conditions["target_size"]
    central_target_radius = conditions["central_target_size"]
    target_index = data["target_indices"][index][i_target]
    target_pos = np.array(data["target_pos"][index][i_target])
    center_pos = np.array([0.0, 0.0])
    to_target_timestamps = np.array(data["to_target_timestamps"][index][i_target])
    to_target_num_timestamps_before_visible = _get_dat(
        data, "to_target_num_timestamps_before_visible", index, i_target, 0
    )
    to_target_mouse_positions = np.stack(
        np.array(data["to_target_mouse_positions"][index][i_target])
    )  # type: ignore
    to_target_success = np.array(data["to_target_success"][index][i_target])
    to_center_success: np.ndarray | bool
    to_center_timestamps = np.array(
        _get_dat(data, "to_center_timestamps", index, i_target, [])
    )
    to_center_num_timestamps_before_visible = _get_dat(
        data, "to_center_num_timestamps_before_visible", index, i_target, 0
    )
    to_center_mouse_positions = np.array(
        _get_dat(data, "to_center_mouse_positions", index, i_target, [])
    )
    if to_center_mouse_positions.shape[0] > 0:
        to_center_mouse_positions = np.stack(to_center_mouse_positions)  # type: ignore
    to_center_success = np.array(
        _get_dat(data, "to_center_success", index, i_target, True)
    )
    return [
        index[0],
        index[1],
        i_target,
        condition_index,
        target_index,
        target_pos,
        target_radius,
        to_target_timestamps,
        to_target_mouse_positions,
        to_target_success,
        to_target_num_timestamps_before_visible,
        center_pos,
        central_target_radius,
        to_center_timestamps,
        to_center_mouse_positions,
        to_center_success,
        to_center_num_timestamps_before_visible,
    ]


def _data_df(trial_handler: TrialHandlerExt) -> pd.DataFrame:
    data = []
    for index in np.ndindex(trial_handler.sequenceIndices.shape):
        # trials that have not yet happened have a default string instead of an array in their data
        if type(trial_handler.data["target_indices"][index]) is np.ndarray:
            n_targets = trial_handler.data["target_indices"][index].shape[0]
            for i_target in range(n_targets):
                data.append(_get_target_data(trial_handler, index, i_target))
    return pd.DataFrame(data, columns=_get_trial_data_columns())


def stats_dataframe(trial_handler: TrialHandlerExt) -> pd.DataFrame:
    df = _data_df(trial_handler)
    for destination in ["target", "center"]:
        df[f"to_{destination}_distance"] = df[f"to_{destination}_mouse_positions"].map(
            _distance
        )
        df[f"to_{destination}_reaction_time"] = df.apply(
            lambda x, destination=destination: _reaction_time(
                x[f"to_{destination}_timestamps"],
                x[f"to_{destination}_mouse_positions"],
                x[f"to_{destination}_num_timestamps_before_visible"],
            ),
            axis=1,
        )
        df[f"to_{destination}_time"] = df.apply(
            lambda x, destination=destination: _total_time(
                x[f"to_{destination}_timestamps"],
                x[f"to_{destination}_num_timestamps_before_visible"],
            ),
            axis=1,
        )
        df[f"to_{destination}_movement_time"] = (
            df[f"to_{destination}_time"] - df[f"to_{destination}_reaction_time"]
        )
        df[f"to_{destination}_rmse"] = df.apply(
            lambda x, destination=destination: _rmse(
                x[f"to_{destination}_mouse_positions"], x[f"{destination}_pos"]
            ),
            axis=1,
        )
        df[f"to_{destination}_spatial_error"] = df.apply(
            lambda x, destination=destination: _spatial_error(
                x[f"to_{destination}_mouse_positions"],
                x[f"{destination}_pos"],
                x[f"{destination}_radius"],
            ),
            axis=1,
        )
    df["area"] = df.apply(
        lambda x: _area(x["to_target_mouse_positions"], x["to_center_mouse_positions"]),
        axis=1,
    )
    df["normalized_area"] = df.apply(
        lambda x: _normalized_area(
            x["to_target_mouse_positions"], x["to_center_mouse_positions"]
        ),
        axis=1,
    )
    df["peak_velocity"] = df.apply(
        lambda x: _peak_velocity(
            np.concatenate((x["to_target_timestamps"], x["to_center_timestamps"])),
            concatenate_mouse_positions(x),
        )[0],
        axis=1,
    )
    df["peak_acceleration"] = df.apply(
        lambda x: _peak_acceleration(
            np.concatenate((x["to_target_timestamps"], x["to_center_timestamps"])),
            concatenate_mouse_positions(x),
        ),
        axis=1,
    )
    df["movement_time_at_peak_velocity"] = df.apply(
        lambda x: _movement_time_at_peak_velocity(
            np.concatenate((x["to_target_timestamps"], x["to_center_timestamps"])),
            concatenate_mouse_positions(x),
            x["to_target_num_timestamps_before_visible"],
        ),
        axis=1,
    )
    df["total_time_at_peak_velocity"] = df.apply(
        lambda x: _total_time_at_peak_velocity(
            np.concatenate((x["to_target_timestamps"], x["to_center_timestamps"])),
            concatenate_mouse_positions(x),
            x["to_target_num_timestamps_before_visible"],
        ),
        axis=1,
    )
    df["movement_distance_at_peak_velocity"] = df.apply(
        lambda x: _movement_distance_at_peak_velocity(
            np.concatenate((x["to_target_timestamps"], x["to_center_timestamps"])),
            concatenate_mouse_positions(x),
            x["to_target_num_timestamps_before_visible"],
        ),
        axis=1,
    )
    df["rmse_movement_at_peak_velocity"] = df.apply(
        lambda x: _rmse_movement_at_peak_velocity(
            np.concatenate((x["to_target_timestamps"], x["to_center_timestamps"])),
            concatenate_mouse_positions(x),
            x["target_pos"],
            x["to_target_num_timestamps_before_visible"],
        ),
        axis=1,
    )
    return df


def concatenate_mouse_positions(x: np.ndarray) -> np.ndarray:
    """
    concatenate the "to_target_mouse_positions" and "to_center_mouse_positions"

    :param x: the data to concatenate
    :return: the concatenated result
    """
    return np.concatenate(
        (
            x["to_target_mouse_positions"],
            x["to_center_mouse_positions"].reshape(
                x["to_center_mouse_positions"].shape[0], 2
            ),
        )
    )


def append_stats_data_to_excel(df: pd.DataFrame, writer: Any, data_format: str) -> None:
    """
    data_format can be
    - "trial": one sheet of data exported per trial
    - "target: one sheet of data exported per target
    """
    if data_format not in ["trial", "target"]:
        raise RuntimeError(f"data_format '{data_format}' not supported")
    data_labels = [
        "to_target_timestamps",
        "to_target_mouse_positions",
        "to_center_timestamps",
        "to_center_mouse_positions",
    ]
    # first sheet: all statistics, exclude arrays of timestamp/position data
    # convert columns of xy values to x column and y column of floats
    df_stats = df.drop(columns=data_labels)
    for label in ["target_pos", "center_pos"]:
        column_as_2d_array = np.stack(df_stats[f"{label}"].to_numpy())
        df_stats[f"{label}_x"] = column_as_2d_array[:, 0]
        df_stats[f"{label}_y"] = column_as_2d_array[:, 1]
        df_stats = df_stats.drop(columns=[label])
    df_stats.to_excel(writer, sheet_name="statistics", index=False)
    # add timestamp/mouse position arrays
    if data_format == "target":
        # one sheet for each row (target) in df, with arrays of time/position data.
        # arrays transposed to columns, x-y pairs are split into two columns.
        # 6 columns:
        #   - to_target_timestamps
        #   - to_center_timestamps
        #   - to_target_mouse_positions_x
        #   - to_target_mouse_positions_y
        #   - to_center_mouse_positions_x
        #   - to_center_mouse_positions_y
        for row in df.itertuples():
            df_data = pd.concat(
                [
                    pd.DataFrame({label: getattr(row, label)})
                    for label in ["to_target_timestamps", "to_center_timestamps"]
                ],
                axis=1,
            )
            for label in ["to_target_mouse_positions", "to_center_mouse_positions"]:
                arr = getattr(row, label)
                if arr.shape[-1] > 0:
                    df_data = pd.concat(
                        [
                            df_data,
                            pd.DataFrame(
                                {f"{label}_x": arr[:, 0], f"{label}_y": arr[:, 1]}
                            ),
                        ],
                        axis=1,
                    )
            df_data.to_excel(writer, sheet_name=f"{row.Index}", index=False)
    else:
        # one sheet per trial: all targets for a trial are concatenated
        # 6 columns:
        #   - timestamps (start from zero, increases throughout the trial)
        #   - mouse_x
        #   - mouse_y
        #   - target_index (-99 if no target is currently displayed)
        #   - target_x (-99 if no target is currently displayed)
        #   - target_y (-99 if no target is currently displayed)
        for i_trial in df.i_trial.unique():
            times = np.array([])
            x_positions = np.array([])
            y_positions = np.array([])
            i_targets = np.array([], dtype=np.int64)
            x_targets = np.array([])
            y_targets = np.array([])
            for row in df.loc[df.i_trial == i_trial].itertuples():
                for dest in ["target", "center"]:
                    raw_times = getattr(row, f"to_{dest}_timestamps")
                    if raw_times.shape[-1] > 0:
                        times = np.append(times, raw_times)
                        points = getattr(row, f"to_{dest}_mouse_positions")
                        x_positions = np.append(x_positions, points[:, 0])
                        y_positions = np.append(y_positions, points[:, 1])
                        # set target info for all timestamps
                        if dest == "center":
                            pos = (0.0, 0.0)
                            i_target = -1  # give center target the special index -1
                            num_timestamps_before_target_visible = (
                                row.to_center_num_timestamps_before_visible
                            )
                        else:
                            pos = row.target_pos
                            i_target = row.i_target
                            num_timestamps_before_target_visible = (
                                row.to_target_num_timestamps_before_visible
                            )
                        target_i = np.full_like(raw_times, i_target, dtype=np.int64)
                        target_x = np.full_like(raw_times, pos[0])
                        target_y = np.full_like(raw_times, pos[1])
                        # if no target is visible use the special index -99
                        target_i[0:num_timestamps_before_target_visible] = -99
                        target_x[0:num_timestamps_before_target_visible] = -99
                        target_y[0:num_timestamps_before_target_visible] = -99
                        i_targets = np.append(i_targets, target_i)
                        x_targets = np.append(x_targets, target_x)
                        y_targets = np.append(y_targets, target_y)
            df_data = pd.DataFrame(
                {
                    "timestamps": times,
                    "mouse_positions_x": x_positions,
                    "mouse_positions_y": y_positions,
                    "i_target": i_targets,
                    "target_x": x_targets,
                    "target_y": y_targets,
                }
            )
            df_data.to_excel(writer, sheet_name=f"{i_trial}", index=False)


def _reaction_time(
    mouse_times: np.ndarray,
    mouse_positions: np.ndarray,
    to_target_num_timestamps_before_visible: int,
) -> float:
    """
    The reaction time is defined as the timestamp where the cursor first moves,
    minus the timestamp where the target becomes visible.
    This means the reaction time can be negative if the cursor is moved before
    the target is made visible.

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :param to_target_num_timestamps_before_visible: The index of the first timestamp where the target is visible
    :return: The reaction time
    """
    if (
        mouse_times.shape[0] != mouse_positions.shape[0]
        or mouse_times.shape[0] == 0
        or mouse_times.shape[0] < to_target_num_timestamps_before_visible
    ):
        return np.nan
    i = 0
    while xydist(mouse_positions[0], mouse_positions[i]) < min_distance and i + 1 < len(
        mouse_times
    ):
        i += 1
    return mouse_times[i] - mouse_times[to_target_num_timestamps_before_visible]


def _total_time(
    mouse_times: np.ndarray,
    to_target_num_timestamps_before_visible: int,
) -> float:
    """
    The time to target is defined as the final timestamp (corresponding either to the target being reached
    or a timeout) minus the timestamp where the target becomes visible.

    :param mouse_times: The array of timestamps
    :param to_target_num_timestamps_before_visible: The index of the first timestamp where the target is visible
    :return: The total time to target
    """
    if (
        mouse_times.shape[0] == 0
        or mouse_times.shape[0] < to_target_num_timestamps_before_visible
    ):
        return np.nan
    return mouse_times[-1] - mouse_times[to_target_num_timestamps_before_visible]


def _distance(mouse_positions: np.ndarray) -> float:
    """
    The euclidean point-to-point distance travelled by the cursor.

    :param mouse_positions: The array of mouse positions
    :return: The distance travelled.
    """
    dist = 0
    for i in range(mouse_positions.shape[0] - 1):
        dist += xydist(mouse_positions[i + 1], mouse_positions[i])
    return dist


def _rmse(mouse_positions: np.ndarray, target_position: np.ndarray) -> float:
    """
    The Root Mean Square Error (RMSE) of the perpendicular distance from each mouse point
    to the straight line that intersects the initial mouse location and the target.

    See: https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#Line_defined_by_two_points.

    :param mouse_positions: The array of mouse positions
    :param target: The x,y coordinates of the target
    :return: The RMSE of the distance from the ideal trajectory
    """
    if mouse_positions.shape[0] <= 1:
        return np.nan
    # use first mouse position as origin point, so exclude it from RMSE measure
    x1, y1 = mouse_positions[0]
    x2, y2 = target_position
    norm = np.power(x2 - x1, 2) + np.power(y2 - y1, 2)
    norm *= mouse_positions.shape[0] - 1
    sum_of_squares = 0
    for x0, y0 in mouse_positions[1:]:
        sum_of_squares += np.power((x2 - x1) * (y1 - y0) - (x1 - x0) * (y2 - y1), 2)
    return np.sqrt(sum_of_squares / norm)


def _area(
    to_target_mouse_positions: np.ndarray, to_center_mouse_positions: np.ndarray
) -> float:
    """
    Calculates the total area enclosed by the mouse positions and the corresponding list of closed polygons

    Uses the built-in operation `area` in the library `shapely` to calculate the area of geometry object.
    However this is only available for valid (not self intersected) geometries.
    To tackle the self-intersection problem,
    the strategy is to split one self intersected object into the union of LineString (a geometry type composed of one or more line segments),
    then construct a bunch of valid polygons from these lines,
    then calculate the area of each valid polygon and sum them up.

    :param to_target_mouse_positions: x,y mouse positions moving towards the target
    :param to_center_mouse_positions: x,y mouse positions moving towards the center
    :return: area

    """
    coords = get_closed_polygon(to_target_mouse_positions, to_center_mouse_positions)
    polygons = polygonize(unary_union(LineString(coords)))
    area = sum(polygon.area for polygon in polygons)
    return area


def _normalized_area(
    to_target_mouse_positions: np.ndarray, to_center_mouse_positions: np.ndarray
) -> float:
    """
    normalized area = (the area formed by paths) / (length of the paths)²

    :param to_target_mouse_positions: x,y mouse positions moving towards the target
    :param to_center_mouse_positions: x,y mouse positions moving towards the center
    :return: normalized area
    """
    area = _area(to_target_mouse_positions, to_center_mouse_positions)
    movement_length = get_movement_length(
        to_target_mouse_positions, to_center_mouse_positions
    )

    normalized_area = area / (movement_length**2) if movement_length != 0 else 0
    return normalized_area


def get_movement_length(
    to_target_mouse_positions: np.ndarray, to_center_mouse_positions: np.ndarray
) -> float:
    """
    calculate the length of the paths connecting the target and the center
    if only 1 path exists, another path is the distance between the head and tail of the existing path.

    :param to_target_mouse_positions: x,y mouse positions moving towards the target
    :param to_center_mouse_positions: x,y mouse positions moving towards the center
    :return: length of the movement
    """
    closed_polygon_coords = get_closed_polygon(
        to_target_mouse_positions, to_center_mouse_positions
    )
    movement_length = _distance(closed_polygon_coords)
    return movement_length


def get_closed_polygon(
    to_target_mouse_positions: np.ndarray, to_center_mouse_positions: np.ndarray
) -> np.ndarray:
    """
    connect the to target path and to center path to a closed polygon

    :param to_target_mouse_positions: x,y mouse positions moving towards the target
    :param to_center_mouse_positions: x,y mouse positions moving towards the center
    :return: x,y mouse positions of the closed polygon

    """
    to_target_mouse_positions = preprocess_mouse_positions(to_target_mouse_positions)
    to_center_mouse_positions = preprocess_mouse_positions(to_center_mouse_positions)
    if to_target_mouse_positions.size == 0:
        return np.concatenate(
            [
                to_center_mouse_positions,
                to_center_mouse_positions[0:1],
            ]
        )
    if to_center_mouse_positions.size == 0:
        return np.concatenate(
            [
                to_target_mouse_positions,
                to_target_mouse_positions[0:1],
            ]
        )

    coords = np.concatenate(
        [
            to_target_mouse_positions,
            to_center_mouse_positions[0:1],
            to_center_mouse_positions,
            to_target_mouse_positions[0:1],
        ]
    )
    return coords


def preprocess_mouse_positions(mouse_positions: np.ndarray) -> np.ndarray:
    """
    reshape the mouse position to (0, 2) if the mouse position is empty, to prevent error happens in the following
    concatenate() function

    :param mouse_positions: x,y mouse positions during movement
    :return: x,y mouse positions after preprocess
    """
    mouse_positions = (
        mouse_positions.reshape(0, 2) if mouse_positions.size == 0 else mouse_positions
    )
    return mouse_positions


def _peak_velocity(
    mouse_times: np.ndarray, mouse_positions: np.ndarray
) -> tuple[np.floating, np.integer]:
    """
    get peak velocity and the corresponding index

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :return: peak velocity and the corresponding index
    """
    velocity = get_velocity(mouse_times, mouse_positions)
    peak_velocity = np.amax(velocity)
    peak_index = np.argmax(velocity)
    return peak_velocity, peak_index


def _peak_acceleration(mouse_times: np.ndarray, mouse_positions: np.ndarray) -> float:
    """
    get peak acceleration

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :return: peak acceleration
    """
    acceleration = get_acceleration(mouse_times, mouse_positions)
    peak_acceleration = np.amax(acceleration)
    return peak_acceleration


def get_derivative(y: np.ndarray, x: np.ndarray) -> np.ndarray:
    """
    get derivative dy/dx

    :param y: the array of y
    :param x: the array of x
    :return: the array of dy/dx
    """
    if x.size <= 1 or y.size <= 1:
        return np.array([0])
    dy_dx = np.diff(y) / np.diff(x)
    return dy_dx


def get_velocity(mouse_times: np.ndarray, mouse_positions: np.ndarray) -> np.ndarray:
    """
    get velocity

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :return: the array of velocity
    """
    first_order_derivative = get_derivative(mouse_positions.transpose(), mouse_times)
    velocity = LA.norm(first_order_derivative, axis=0)
    return velocity


def get_acceleration(
    mouse_times: np.ndarray, mouse_positions: np.ndarray
) -> np.ndarray:
    """
    get acceleration

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :return: the array of acceleration
    """
    first_order_derivative = get_derivative(mouse_positions.transpose(), mouse_times)
    second_order_derivative = get_derivative(first_order_derivative, mouse_times[:-1])
    acceleration = LA.norm(second_order_derivative, axis=0)
    return acceleration


def _spatial_error(
    mouse_position: np.ndarray, target: np.ndarray, target_radius: float
) -> float:
    """
    at timeout linear distance from cursor to target - target radius

    :param mouse_position: The array of mouse positions
    :param target: The position of the target
    :param target_radius: radius of the target
    :return: the distance between the end point of the movement to the center of the target - target radius
    """
    if mouse_position.size < 1:
        return 0
    spatial_error = xydist(mouse_position[-1], target) - target_radius
    return max(spatial_error, 0)


def get_first_movement_index(
    mouse_times: np.ndarray,
    mouse_positions: np.ndarray,
    to_target_num_timestamps_before_visible: int,
) -> int | None:
    """
    get index of the first movement in mouse_times

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :param to_target_num_timestamps_before_visible: The index of the first timestamp where the target is visible
    :return: the first movement index in mouse_times
    """
    if (
        mouse_times.shape[0] != mouse_positions.shape[0]
        or mouse_times.shape[0] == 0
        or mouse_times.shape[0] < to_target_num_timestamps_before_visible
    ):
        return None
    i = 0
    while xydist(mouse_positions[0], mouse_positions[i]) < min_distance and i + 1 < len(
        mouse_times
    ):
        i += 1
    return i


def _movement_time_at_peak_velocity(
    mouse_times: np.ndarray,
    mouse_positions: np.ndarray,
    to_target_num_timestamps_before_visible: int,
) -> float:
    """
    get the time from first movement to the peak velocity

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :param to_target_num_timestamps_before_visible: The index of the first timestamp where the target is visible
    :return: the time from first movement to the peak velocity
    """
    i = get_first_movement_index(
        mouse_times, mouse_positions, to_target_num_timestamps_before_visible
    )
    _, peak_index = _peak_velocity(mouse_times, mouse_positions)
    return (
        mouse_times[peak_index] - mouse_times[i]
        if i is not None and peak_index >= i
        else np.nan
    )


def _total_time_at_peak_velocity(
    mouse_times: np.ndarray,
    mouse_positions: np.ndarray,
    to_target_num_timestamps_before_visible: int,
) -> float:
    """
    get the time from the target becomes visible to the peak velocity

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :param to_target_num_timestamps_before_visible: The index of the first timestamp where the target is visible
    :return: the time from the target becomes visible to the peak velocity
    """
    _, peak_index = _peak_velocity(mouse_times, mouse_positions)
    return (
        mouse_times[peak_index] - mouse_times[to_target_num_timestamps_before_visible]
        if to_target_num_timestamps_before_visible < peak_index
        else np.nan
    )


def _movement_distance_at_peak_velocity(
    mouse_times: np.ndarray,
    mouse_positions: np.ndarray,
    to_target_num_timestamps_before_visible: int,
) -> float:
    """
    get the euclidean point-to-point distance travelled from first movement to the peak velocity

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :param to_target_num_timestamps_before_visible: The index of the first timestamp where the target is visible
    :return: the distance travelled from first movement -> the peak velocity
    """
    i = get_first_movement_index(
        mouse_times, mouse_positions, to_target_num_timestamps_before_visible
    )
    _, peak_index = _peak_velocity(mouse_times, mouse_positions)
    return (
        _distance(mouse_positions[i : peak_index + 1])
        if i is not None and peak_index >= i
        else np.nan
    )


def _rmse_movement_at_peak_velocity(
    mouse_times: np.ndarray,
    mouse_positions: np.ndarray,
    target_position: np.ndarray,
    to_target_num_timestamps_before_visible: int,
) -> float:
    """
    The Root Mean Square Error (RMSE) of the perpendicular distance from the peak velocity mouse point
    to the straight line that intersects the first mouse location and the target.

    See: https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#Line_defined_by_two_points.

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :param target_position: The position of the target
    :param to_target_num_timestamps_before_visible: The index of the first timestamp where the target is visible
    :return: The RMSE of the distance from the ideal trajectory
    """
    i = get_first_movement_index(
        mouse_times, mouse_positions, to_target_num_timestamps_before_visible
    )
    if i is not None:
        p1 = mouse_positions[i]
    else:
        return np.nan
    p2 = target_position
    _, peak_index = _peak_velocity(mouse_times, mouse_positions)
    p3 = mouse_positions[peak_index]
    return (
        float(LA.norm(np.cross(p2 - p1, p1 - p3)) / LA.norm(p2 - p1))
        if peak_index >= i
        else np.nan
    )



from __future__ import annotations

import sys

from pytest import MonkeyPatch

import vstt
from vstt.update import check_for_new_version
from vstt.update import do_pip_upgrade


def test_new_version_available(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(vstt, "__version__", "0.0.0")
    available, message = check_for_new_version()
    assert available is True
    assert "0.0.0" in message
    assert "upgrade" in message
    monkeypatch.setattr(vstt, "__version__", "9999.0.0")
    available, message = check_for_new_version()
    assert available is False
    assert "latest version" in message
    monkeypatch.setattr(vstt, "__version__", "imnot-a-valid-version!")
    available, message = check_for_new_version()
    assert available is False
    assert "error" in message


def test_do_pip_upgrade(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(sys, "executable", "abc123_i_dont_exist")
    success, message = do_pip_upgrade()
    assert success is False
    assert "error" in message



from __future__ import annotations

import pathlib
from typing import Any

import gui_test_utils as gtu
import numpy as np
import qt_test_utils as qtu
from PIL import Image
from psychopy.visual.window import Window
from pytest import MonkeyPatch
from qtpy.QtWidgets import QFileDialog

from vstt.experiment import Experiment
from vstt.results_widget import ResultsWidget


def test_results_widget_no_experiment(window: Window) -> None:
    widget = ResultsWidget(parent=None, win=window)
    # initially empty
    assert widget.experiment.trial_handler_with_results is None
    assert widget._list_trials.count() == 0
    assert widget._btn_display_trial.isEnabled() is False
    assert widget._btn_screenshot_trial.isEnabled() is False
    assert widget._btn_display_condition.isEnabled() is False
    assert widget._btn_screenshot_condition.isEnabled() is False


def test_results_widget_experiment_no_results(
    experiment_no_results: Experiment, window: Window
) -> None:
    widget = ResultsWidget(parent=None, win=window)
    # assign experiment without results
    widget.experiment = experiment_no_results
    assert widget.experiment.trial_handler_with_results is None
    assert widget._list_trials.count() == 0
    assert widget._btn_display_trial.isEnabled() is False
    assert widget._btn_screenshot_trial.isEnabled() is False
    assert widget._btn_display_condition.isEnabled() is False
    assert widget._btn_screenshot_condition.isEnabled() is False


def test_results_widget_experiment_with_results(
    experiment_with_results: Experiment,
    window: Window,
    tmp_path: pathlib.Path,
    monkeypatch: MonkeyPatch,
) -> None:
    png_path = tmp_path / "tmp.png"

    # monkey patch QFileDialog.getSaveFileName to just return desired filename
    def mock_get_save_file_name(*args: Any, **kwargs: Any) -> tuple[str, str]:
        return str(png_path), ""

    monkeypatch.setattr(QFileDialog, "getSaveFileName", mock_get_save_file_name)

    widget = ResultsWidget(parent=None, win=window)
    # assign experiment with results
    widget.experiment = experiment_with_results
    n_trials = 4
    assert widget.experiment.trial_handler_with_results is not None
    assert widget.experiment.trial_handler_with_results.nTotal == n_trials
    assert widget._list_trials.count() == n_trials
    # select first row
    for _ in range(n_trials):
        qtu.press_up_key(widget._list_trials)
    assert widget._list_trials.currentRow() == 0
    # display trial results
    screenshot = gtu.call_target_and_get_screenshot(
        qtu.click,
        (widget._btn_display_trial,),
        window,
    )
    # most pixels grey
    trial_grey_frac = gtu.pixel_color_fraction(screenshot, (128, 128, 128))
    assert 0.900 < trial_grey_frac < 0.999
    # some off-white pixels
    assert 0.001 <= gtu.pixel_color_fraction(screenshot, (240, 248, 255)) <= 0.050
    # save image of trial results
    qtu.click(widget._btn_screenshot_trial)
    assert png_path.is_file()
    img = Image.open(png_path)
    saved_image = np.asarray(img)
    assert np.allclose(saved_image.shape, screenshot.shape)
    # display condition results
    screenshot = gtu.call_target_and_get_screenshot(
        qtu.click,
        (widget._btn_display_condition,),
        window,
    )
    # less grey pixels than a single trial
    assert gtu.pixel_color_fraction(screenshot, (128, 128, 128)) < trial_grey_frac
    # some off-white pixels
    assert 0.001 <= gtu.pixel_color_fraction(screenshot, (240, 248, 255)) <= 0.050
    # save image of condition results
    qtu.click(widget._btn_screenshot_condition)
    assert png_path.is_file()
    img = Image.open(png_path)
    saved_image = np.asarray(img)
    assert np.allclose(saved_image.shape, screenshot.shape)
    # select each row in turn
    for row in range(n_trials):
        assert widget._list_trials.currentRow() == row
        assert widget._btn_display_trial.isEnabled() is True
        assert widget._btn_display_condition.isEnabled() is True
        qtu.press_down_key(widget._list_trials)
    assert widget._list_trials.currentRow() == n_trials - 1
    # assign a new experiment without results
    experiment_with_results.clear_results()
    widget.experiment = experiment_with_results
    assert widget.experiment.trial_handler_with_results is None
    assert widget._list_trials.count() == 0
    assert widget._btn_display_trial.isEnabled() is False
    assert widget._btn_display_condition.isEnabled() is False


def test_results_widget_experiment_invalid_state(
    experiment_with_results: Experiment, window: Window
) -> None:
    widget = ResultsWidget(parent=None, win=window)
    # assign experiment with results
    widget.experiment = experiment_with_results
    n_trials = 4
    assert widget.experiment.trial_handler_with_results is not None
    assert widget.experiment.trial_handler_with_results.nTotal == n_trials
    assert widget._list_trials.count() == n_trials
    # select first row
    for _ in range(n_trials):
        qtu.press_up_key(widget._list_trials)
    assert widget._list_trials.count() == n_trials
    assert widget._list_trials.currentRow() == 0
    assert widget._btn_display_trial.isEnabled() is True
    assert widget._btn_display_condition.isEnabled() is True
    # externally delete results from experiment without updating the widget
    th_with_results = experiment_with_results.trial_handler_with_results
    experiment_with_results.clear_results()
    assert widget.experiment.trial_handler_with_results is None
    # widget state is now invalid: still have enabled buttons but no results
    assert widget._list_trials.count() == n_trials
    assert widget._list_trials.currentRow() == 0
    assert widget._btn_display_trial.isEnabled() is True
    assert widget._btn_display_condition.isEnabled() is True
    # click on display_trial button: widget checks for invalid state and updates to match the experiment
    qtu.click(widget._btn_display_trial)
    assert widget._list_trials.count() == 0
    assert widget._btn_display_trial.isEnabled() is False
    assert widget._btn_display_condition.isEnabled() is False
    # repeat for display_condition button
    assert th_with_results is not None
    # restore results to experiment
    experiment_with_results.trial_handler_with_results = th_with_results
    widget.experiment = experiment_with_results
    assert widget._list_trials.count() == n_trials
    # select first row
    for _ in range(n_trials):
        qtu.press_up_key(widget._list_trials)
    assert widget._list_trials.currentRow() == 0
    # clear results without updating widget
    experiment_with_results.clear_results()
    assert widget._btn_display_trial.isEnabled() is True
    assert widget._btn_display_condition.isEnabled() is True
    # click on display trial button: widget updates invalid state
    qtu.click(widget._btn_display_condition)
    assert widget._list_trials.count() == 0
    assert widget._btn_display_trial.isEnabled() is False
    assert widget._btn_display_condition.isEnabled() is False



from __future__ import annotations

import pytest

import vstt


def test_display_options_labels() -> None:
    display_options = vstt.display.default_display_options()
    labels = vstt.display.display_options_labels()
    assert len(display_options) == len(labels)
    assert display_options.keys() == labels.keys()


def test_import_display_options(caplog: pytest.LogCaptureFixture) -> None:
    default_display_options = vstt.display.default_display_options()
    display_options_dict = {
        "to_target_paths": True,
        "to_center_paths": "I should be a bool!",
        "targets": True,
        "central_target": True,
        "to_target_reaction_time": True,
        "to_center_reaction_time": True,
        "to_target_movement_time": True,
        "to_center_movement_time": True,
        "to_target_time": True,
        "to_center_time": True,
        "to_target_distance": False,
        "to_center_distance": False,
        "to_target_rmse": False,
        "to_center_rmse": False,
        "averages": True,
        "unknown_key1": "ignore me",
        "unknown_key2": False,
        "to_target_success": False,
        "to_center_success": False,
        "area": False,
        "normalized_area": False,
        "peak_velocity": False,
        "peak_acceleration": False,
        "to_target_spatial_error": False,
        "to_center_spatial_error": False,
        "movement_time_at_peak_velocity": False,
        "total_time_at_peak_velocity": False,
        "movement_distance_at_peak_velocity": False,
        "rmse_movement_at_peak_velocity": False,
    }
    for key in default_display_options:
        assert key in display_options_dict
    # if any keys are missing or have invalid values, default values are used instead
    missing_keys = [
        "to_center_reaction_time",
        "to_target_rmse",
        "to_center_rmse",
    ]
    for key in missing_keys:
        display_options_dict.pop(key)
    display_options = vstt.display.import_display_options(display_options_dict)
    for key, value in display_options.items():
        if key in missing_keys:
            assert value == default_display_options[key]  # type: ignore
        else:
            value_type = type(value)
            assert value == value_type(display_options_dict[key])  # type: ignore
    log_messages = {r.message for r in caplog.records}
    expected_log_messages = {
        "Key 'to_center_reaction_time' missing, using default 'False'",
        "Key 'to_target_rmse' missing, using default 'True'",
        "Key 'to_center_rmse' missing, using default 'False'",
        "Key 'to_center_paths' invalid: expected <class 'bool'>, got <class 'str'> 'I should be a bool!'",
        "Ignoring unknown key 'unknown_key1'",
        "Ignoring unknown key 'unknown_key2'",
    }
    assert log_messages == expected_log_messages
